// Thread block size
#define TILE_SIZE 16

// OpenCL Kernel for matrix multiply, C = A * B
// with boundary conditions
__kernel void matrixMul(__global float* C,
                        __global float* A,
                        __global float* B,
                        int m,
                        int k,
                        int n) {

  int tx = get_global_id(0); //2D Global Thread ID x
  int ty = get_global_id(1); //2D Global Thread ID y

  if ((tx < n) && (ty < m)) {
    // Initialize accumulator
    float res = 0.0f;

    // Perform dot-product of row-column
    for (int i = 0; i < k; i++) {
      res += A[ty*k+i] * B[i*n+tx];
    }

    // Write in device memory
    C[ty*n+tx] = res;
  }

}

