/***************************************************************************
 *cr
 *cr            (C) Copyright 2010 The Board of Trustees of the
 *cr                        University of Illinois
 *cr                         All Rights Reserved
 *cr
 ***************************************************************************/

/*
 * Main entry of dense matrix-matrix multiplication kernel
 */

#include "opencv2/opencv.hpp"
#include "opencv2/core/ocl.hpp"
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <malloc.h>
#include <iostream>
#include <cassert>
#include <parboil.h>
#include <visc.h>


#define NUM_RUNS 1 
#define DEPTH 3
#define HEIGHT 640
#define WIDTH 480
std::string input_window = "GPU Pipeline - Input Video";
std::string output_window = "GPU Pipeline - Edge Mapping";


#ifdef MIDDLE
  #define POSX_IN     640
  #define POSY_IN     0
  #define POSX_OUT    640
  #define POSY_OUT    540

#elif RIGHT
  #define POSX_IN     1280
  #define POSY_IN     0
  #define POSX_OUT    1280
  #define POSY_OUT    540

#else // LEFT
  #define POSX_IN     0
  #define POSY_IN     0
  #define POSX_OUT    0
  #define POSY_OUT    540
#endif


//#define NUM_FRAMES 20



// Definitions of sizes for edge detection kernels

#define MIN_BR 0.0f
#define MAX_BR 1.0f

// Code needs to be changed for this to vary
#define SZB 3

#define REDUCTION_TILE_SZ 1024

#define _MIN(X,Y) ((X) < (Y) ? (X) : (Y))
#define _MAX(X,Y) ((X) > (Y) ? (X) : (Y))

extern "C" {

struct __attribute__((__packed__)) InStruct {
  float* I ;
  size_t bytesI;
  float* B;
  size_t bytesB;
  float* D;
  size_t bytesD;
  float* E;
  size_t bytesE;
  float* L;
  size_t bytesL;
  long m;
  long n;
};


void packData(struct InStruct* args, float* I, size_t bytesI,
                                     float* B, size_t bytesB,
                                     float* D, size_t bytesD,
                                     float* E, size_t bytesE,
                                     float* L, size_t bytesL,
                                     long m, long n) {
  args->I = I;
  args->bytesI = bytesI;
  args->B = B;
  args->bytesB = bytesB;
  args->D = D;
  args->bytesD = bytesD;
  args->E = E;
  args->bytesE = bytesE;
  args->L = L;
  args->bytesL = bytesL;
  args->m = m;
  args->n = n;
}

void dilate(float* Is, size_t bytesIs,
            float* B, size_t bytesB,
            float* D, size_t bytesD,
            long m, long n) {

  __visc__hint(visc::DEVICE);
  __visc__attributes(2, Is, B, 1, D);


  // 3x3 image area
  float imageArea[SZB][SZB];

  //int gx = get_global_id(0);
  //int gy = get_global_id(1);
  void* thisNode = __visc__getNode();
  long gx = __visc__getNodeInstanceID_x(thisNode);
  long gy = __visc__getNodeInstanceID_y(thisNode);
  //if(gx == 0 && gy == 0)
    //std::cout << "Entered laplacian\n";

  int i, j;

  if ((gx < n) && (gy < m)) {
    // Data copy for dilation filter
    imageArea[1][1] = Is[gy * n + gx];

    if (gx == 0) {
      imageArea[0][0] = imageArea[1][0] = imageArea[2][0] = MIN_BR;
    } else {
      imageArea[1][0] = Is[gy * n + gx - 1];
      imageArea[0][0] = (gy > 0) ? Is[(gy - 1) * n + gx - 1] : MIN_BR;
      imageArea[2][0] = (gy < m - 1) ? Is[(gy + 1) * n + gx - 1] : MIN_BR;
    }

    if (gx == n - 1) {
      imageArea[0][2] = imageArea[1][2] = imageArea[2][2] = MIN_BR;
    } else {
      imageArea[1][2] = Is[gy * n + gx + 1];
      imageArea[0][2] = (gy > 0) ? Is[(gy - 1) * n + gx + 1] : MIN_BR;
      imageArea[2][2] = (gy < m - 1) ? Is[(gy + 1) * n + gx + 1] : MIN_BR;
    }

    imageArea[0][1] = (gy > 0) ? Is[(gy - 1) * n + gx] : MIN_BR;
    imageArea[2][1] = (gy < m - 1) ? Is[(gy + 1) * n + gx] : MIN_BR;

    // Compute pixel of dilated image
    float dilatedPixel = MIN_BR;
    for (i = 0; i < SZB; i++)
      for (j = 0; j < SZB; j++)
        dilatedPixel = _MAX(dilatedPixel, imageArea[i][j] * B[i*SZB + j]);

    D[gy*n+gx] = dilatedPixel;
  }
  //__visc__return(m);
  __visc__return(bytesB);
}

void WrapperDilate(float* Is, size_t bytesIs,
            float* B, size_t bytesB,
            float* D, size_t bytesD,
            long m, long n) {

  __visc__hint(visc::CPU_TARGET);
  __visc__attributes(2, Is, B, 1, D);

  void* Node = __visc__createNode2D(dilate, m, n);
  __visc__bindIn(Node, 0, 0, 0); // Bind Is
  __visc__bindIn(Node, 1, 1, 0); // Bind bytesIs
  __visc__bindIn(Node, 2, 2, 0); // Bind B
  __visc__bindIn(Node, 3, 3, 0); // Bind bytesB
  __visc__bindIn(Node, 4, 4, 0); // Bind D
  __visc__bindIn(Node, 5, 5, 0); // Bind bytesD
  __visc__bindIn(Node, 6, 6, 0); // Bind m
  __visc__bindIn(Node, 7, 7, 0); // Bind n

  __visc__bindOut(Node, 0, 0, 0); // bind output m


}

void erode( float* Is, size_t bytesIs,
            float* B, size_t bytesB,
            float* E, size_t bytesE,
            long m, long n) {

  __visc__hint(visc::DEVICE);
  __visc__attributes(2, Is, B, 1, E);

  // 3x3 image area
  float imageArea[SZB][SZB];

  //int gx = get_global_id(0);
  //int gy = get_global_id(1);
  void* thisNode = __visc__getNode();
  long gx = __visc__getNodeInstanceID_x(thisNode);
  long gy = __visc__getNodeInstanceID_y(thisNode);
  //if(gx == 0 && gy == 0)
    //std::cout << "Entered laplacian\n";

  int i, j;

  if ((gx < n) && (gy < m)) {
    // Data copy for dilation filter
    imageArea[1][1] = Is[gy * n + gx];

    if (gx == 0) {
      imageArea[0][0] = imageArea[1][0] = imageArea[2][0] = MAX_BR;
    } else {
      imageArea[1][0] = Is[gy * n + gx - 1];
      imageArea[0][0] = (gy > 0) ? Is[(gy - 1) * n + gx - 1] : MAX_BR;
      imageArea[2][0] = (gy < m - 1) ? Is[(gy + 1) * n + gx - 1] : MAX_BR;
    }

    if (gx == n - 1) {
      imageArea[0][2] = imageArea[1][2] = imageArea[2][2] = MAX_BR;
    } else {
      imageArea[1][2] = Is[gy * n + gx + 1];
      imageArea[0][2] = (gy > 0) ? Is[(gy - 1) * n + gx + 1] : MAX_BR;
      imageArea[2][2] = (gy < m - 1) ? Is[(gy + 1) * n + gx + 1] : MAX_BR;
    }

    imageArea[0][1] = (gy > 0) ? Is[(gy - 1) * n + gx] : MAX_BR;
    imageArea[2][1] = (gy < m - 1) ? Is[(gy + 1) * n + gx] : MAX_BR;

    // Compute pixel of dilated image
    float erodedPixel = MIN_BR;
    for (i = 0; i < SZB; i++)
      for (j = 0; j < SZB; j++)
        erodedPixel = _MIN(erodedPixel, imageArea[i][j] * B[i*SZB + j]);

    E[gy*n+gx] = erodedPixel;
  }
  //__visc__return(n);
  __visc__return(bytesB);
 
}

void WrapperErode(float* Is, size_t bytesIs,
            float* B, size_t bytesB,
            float* E, size_t bytesE,
            long m, long n) {

  __visc__hint(visc::CPU_TARGET);
  __visc__attributes(2, Is, B, 1, E);

  void* Node = __visc__createNode2D(erode, m, n);
  __visc__bindIn(Node, 0, 0, 0); // Bind Is
  __visc__bindIn(Node, 1, 1, 0); // Bind bytesIs
  __visc__bindIn(Node, 2, 2, 0); // Bind B
  __visc__bindIn(Node, 3, 3, 0); // Bind bytesB
  __visc__bindIn(Node, 4, 4, 0); // Bind E
  __visc__bindIn(Node, 5, 5, 0); // Bind bytesE
  __visc__bindIn(Node, 6, 6, 0); // Bind m
  __visc__bindIn(Node, 7, 7, 0); // Bind n

  __visc__bindOut(Node, 0, 0, 0); // bind output n


}

void lincomb( float* Is, size_t bytesIs,
              float* D, size_t bytesD,
              float* E, size_t bytesE,
              float* L, size_t bytesL,
              long m, long n,
              size_t dummy1, size_t dummy2) {

  __visc__hint(visc::DEVICE);
  __visc__attributes(3, Is, E, D, 1, L);

  void* thisNode = __visc__getNode();
  long gx = __visc__getNodeInstanceID_x(thisNode);
  long gy = __visc__getNodeInstanceID_y(thisNode);
  
  if ((gx < n) && (gy < m)) {
    long index = gy*n+gx;
    L[index] = D[index] + E[index] - 2 * Is[index];
  }

  //__visc__return(m);
}

void WrapperLincomb(float* Is, size_t bytesIs,
            float* D, size_t bytesD,
            float* E, size_t bytesE,
            float* L, size_t bytesL,
            long m, long n,
            size_t dummy1, size_t dummy2) {

  __visc__hint(visc::CPU_TARGET);
  __visc__attributes(3, Is, E, D, 1, L);

  void* Node = __visc__createNode2D(lincomb, m, n);
  __visc__bindIn(Node, 0, 0, 0); // Bind Is
  __visc__bindIn(Node, 1, 1, 0); // Bind bytesIs
  __visc__bindIn(Node, 2, 2, 0); // Bind D
  __visc__bindIn(Node, 3, 3, 0); // Bind bytesD
  __visc__bindIn(Node, 4, 4, 0); // Bind E
  __visc__bindIn(Node, 5, 5, 0); // Bind bytesE
  __visc__bindIn(Node, 6, 6, 0); // Bind L
  __visc__bindIn(Node, 7, 7, 0); // Bind bytesL
  __visc__bindIn(Node, 8, 8, 0); // Bind m
  __visc__bindIn(Node, 9, 9, 0); // Bind n
  __visc__bindIn(Node, 10, 10, 0); // Bind dummy1
  __visc__bindIn(Node, 11, 11, 0); // Bind dummy2

  //__visc__bindOut(Node, 0, 0, 0); // bind output m


}

/* Compute a non-linear laplacian estimate of input image I of size m x n */
/*
 * Is   : blurred imput image
 * m, n : dimensions
 * B    : structural element for dilation - erosion ([0 1 0; 1 1 1; 0 1 0])
 * L    : output (laplacian of the image)
 * Need 2D grid, a thread per pixel
*/

/*
void laplacianEstimate(float *Is, size_t bytesIs,
                            float *B, size_t bytesB,
                            float *L, size_t bytesL,
                            int m, int n) {

  __visc__hint(visc::DEVICE);
  __visc__attributes(2, Is, B, 1, L);
  // 3x3 image area
  float imageArea[SZB][SZB];

  //int gx = get_global_id(0);
  //int gy = get_global_id(1);
  void* thisNode = __visc__getNode();
  int gx = __visc__getNodeInstanceID_x(thisNode);
  int gy = __visc__getNodeInstanceID_y(thisNode);
  //if(gx == 0 && gy == 0)
    //std::cout << "Entered laplacian\n";

  int i, j;

  if ((gx < n) && (gy < m)) {
    // Data copy for dilation filter
    imageArea[1][1] = Is[gy * n + gx];

    if (gx == 0) {
      imageArea[0][0] = imageArea[1][0] = imageArea[2][0] = MIN_BR;
    } else {
      imageArea[1][0] = Is[gy * n + gx - 1];
      imageArea[0][0] = (gy > 0) ? Is[(gy - 1) * n + gx - 1] : MIN_BR;
      imageArea[2][0] = (gy < m - 1) ? Is[(gy + 1) * n + gx - 1] : MIN_BR;
    }

    if (gx == n - 1) {
      imageArea[0][2] = imageArea[1][2] = imageArea[2][2] = MIN_BR;
    } else {
      imageArea[1][2] = Is[gy * n + gx + 1];
      imageArea[0][2] = (gy > 0) ? Is[(gy - 1) * n + gx + 1] : MIN_BR;
      imageArea[2][2] = (gy < m - 1) ? Is[(gy + 1) * n + gx + 1] : MIN_BR;
    }

    imageArea[0][1] = (gy > 0) ? Is[(gy - 1) * n + gx] : MIN_BR;
    imageArea[2][1] = (gy < m - 1) ? Is[(gy + 1) * n + gx] : MIN_BR;

    // Compute pixel of dilated image
    float dilatedPixel = MIN_BR;
    for (i = 0; i < SZB; i++)
      for (j = 0; j < SZB; j++)
        dilatedPixel = _MAX(dilatedPixel, imageArea[i][j] * B[i*SZB + j]);

    // Data copy for erotion filter - only change the boundary conditions
    if (gx == 0) {
      imageArea[0][0] = imageArea[1][0] = imageArea[2][0] = MAX_BR;
    } else {
      if (gy == 0) imageArea[0][0] = MAX_BR;
      if (gy == m-1) imageArea[2][0] = MAX_BR;
    }

    if (gx == n - 1) {
      imageArea[0][2] = imageArea[1][2] = imageArea[2][2] = MAX_BR;
    } else {
      if (gy == 0) imageArea[0][2] = MAX_BR;
      if (gy == m-1) imageArea[2][2] = MAX_BR;
    }

    if (gy == 0) imageArea[0][1] = MAX_BR;
    if (gy == m-1) imageArea[2][1] = MAX_BR;

    // Compute pixel of eroded image
    float erodedPixel = MAX_BR;
    for (i = 0; i < SZB; i++)
      for (j = 0; j < SZB; j++)
        erodedPixel = _MIN(erodedPixel, imageArea[i][j] * B[i*SZB + j]);

    float laplacian = dilatedPixel + erodedPixel - 2 * imageArea[1][1];
    L[gy*n+gx] = laplacian;
  }
  //OutStruct output = {bytesB, bytesL};
  //if(gx == m-1 && gy == n-1)
    //std::cout << "Exit laplacian\n";
  __visc__return(m);
}

void WrapperlaplacianEstimate(float *Is, size_t bytesIs,
                          float *B, size_t bytesB,
                          float *L, size_t bytesL,
                          int m, int n) {
  __visc__hint(visc::CPU_TARGET);
  __visc__attributes(2, Is, B, 1, L);
  void* LNode = __visc__createNode2D(laplacianEstimate, m, n);
  __visc__bindIn(LNode, 0, 0, 0); // Bind Is
  __visc__bindIn(LNode, 1, 1, 0); // Bind bytesIs
  __visc__bindIn(LNode, 2, 2, 0); // Bind B
  __visc__bindIn(LNode, 3, 3, 0); // Bind bytesB
  __visc__bindIn(LNode, 4, 4, 0); // Bind L
  __visc__bindIn(LNode, 5, 5, 0); // Bind bytesL
  __visc__bindIn(LNode, 6, 6, 0); // Bind m
  __visc__bindIn(LNode, 7, 7, 0); // Bind n

  __visc__bindOut(LNode, 0, 0, 0); // bind output m

}
*/

void LaplacianEstimate(float* Is, size_t bytesIs,// 0
                          float* B, size_t bytesB,      // 2
                          float* D, size_t bytesD,      // 4
                          float* E, size_t bytesE,      // 6
                          float* L, size_t bytesL,      // 8
                          long m, long n) {             // 10
  __visc__hint(visc::CPU_TARGET);
  __visc__attributes(2, Is, B, 3, D, E, L);
  void* DNode = __visc__createNode(WrapperDilate);
  void* ENode = __visc__createNode(WrapperErode);
  void* LNode = __visc__createNode(WrapperLincomb);

  // Dilate
  __visc__bindIn(DNode, 0, 0, 0); // Bind Is
  __visc__bindIn(DNode, 1, 1, 0); // Bind bytesIs
  __visc__bindIn(DNode, 2, 2, 0); // Bind B
  __visc__bindIn(DNode, 3, 3, 0); // Bind bytesB
  __visc__bindIn(DNode, 4, 4, 0); // Bind D
  __visc__bindIn(DNode, 5, 5, 0); // Bind bytesD
  __visc__bindIn(DNode, 10, 6, 0); // Bind m
  __visc__bindIn(DNode, 11, 7, 0); // Bind n

  // Erode
  __visc__bindIn(ENode, 0, 0, 0); // Bind Is
  __visc__bindIn(ENode, 1, 1, 0); // Bind bytesIs
  __visc__bindIn(ENode, 2, 2, 0); // Bind B
  __visc__bindIn(ENode, 3, 3, 0); // Bind bytesB
  __visc__bindIn(ENode, 6, 4, 0); // Bind E
  __visc__bindIn(ENode, 7, 5, 0); // Bind bytesE
  __visc__bindIn(ENode, 10, 6, 0); // Bind m
  __visc__bindIn(ENode, 11, 7, 0); // Bind n


  // Linear combination
  __visc__bindIn(LNode, 0, 0, 0); // Bind Is
  __visc__bindIn(LNode, 1, 1, 0); // Bind bytesIs
  __visc__bindIn(LNode, 4, 2, 0); // Bind D
  __visc__bindIn(LNode, 5, 3, 0); // Bind bytesD
  __visc__bindIn(LNode, 6, 4, 0); // Bind E
  __visc__bindIn(LNode, 7, 5, 0); // Bind bytesE
  __visc__bindIn(LNode, 8, 6, 0); // Bind L
  __visc__bindIn(LNode, 9, 7, 0); // Bind bytesL
  __visc__bindIn(LNode, 10, 8, 0); // Bind m [Alt: Pass as edge from dilate]
  __visc__bindIn(LNode, 11, 9, 0); // Bind n [Alt: Pass as edge from erode]

  //__visc__edge(DNode, LNode, 0, 0, 8, 0); // Get m from dilate
  //__visc__edge(ENode, LNode, 0, 0, 9, 0); // Get n from erode
  __visc__edge(DNode, LNode, 0, 0, 10, 0); // Get a dummy from dilate
  __visc__edge(ENode, LNode, 0, 0, 11, 0); // Get a dummy from erode

  //__visc__bindOut(LNode, 0, 0, 0); // bind output m

}

}
using namespace cv;

void getNextFrame(VideoCapture& VC, Mat& F) {
  VC >> F;
  /// Convert the image to grayscale if image colored
  if(F.channels() == 3)
    cvtColor( F, F, CV_BGR2GRAY );

  F.convertTo(F, CV_32F, 1.0/255.0);

}

//void showInOut(Mat& Input, Mat& Output) {
    //Mat in, out;
    //resize(Input, in, Size(512, 768));
    //resize(Output, out, Size(512, 768));
    //imshow(input_window, in);
    //imshow(output_window, out);
//}


int main (int argc, char *argv[]) {

    struct pb_Parameters *params;
    struct pb_TimerSet timers;

    size_t I_sz;

    std::cout << "Using OpenCV" << CV_VERSION << "\n";

    /* Read command line. Expect 3 inputs: A, B and B^T
       in column-major layout*/
    params = pb_ReadParameters(&argc, argv);
    if ((params->inpFiles[0] == NULL)
            || (params->inpFiles[1] != NULL))
    {
        fprintf(stderr, "Expecting input image filename\n");
        exit(-1);
    }

    /* Read in data */
    std::cout << "Reading video file: " << params->inpFiles[0] << "\n";
    VideoCapture cap(params->inpFiles[0]);
    if(!cap.isOpened()) {
      std::cout << "Could not open video file" << "\n";
      return -1;
    }

    int NUM_FRAMES = cap.get(CV_CAP_PROP_FRAME_COUNT);
    std::cout << "Number of frames = " << NUM_FRAMES << "\n";
    namedWindow(input_window, CV_WINDOW_AUTOSIZE);
    namedWindow(output_window, CV_WINDOW_AUTOSIZE);
    moveWindow(input_window, POSX_IN, POSY_IN);
    moveWindow(output_window, POSX_OUT, POSY_OUT);

    Mat src, D, E, L;

    getNextFrame(cap, src);

    std::cout << "Image dimension = " << src.size() << "\n";
    if(!src.isContinuous()) {
      std::cout << "Expecting contiguous storage of image in memory!\n";
      exit(-1);
    }

    D = Mat(src.size[0], src.size[1], CV_32F);
    E = Mat(src.size[0], src.size[1], CV_32F);
    L = Mat(src.size[0], src.size[1], CV_32F);

    // All these matrices need to have their data array contiguous in memory
    assert(src.isContinuous() && L.isContinuous() && D.isContinuous() && E.isContinuous());

    pb_InitializeTimerSet(&timers);
    __visc__init();

    //pb_SwitchToTimer( &timers, pb_TimerID_COMPUTE );
    // copy A to device memory
    I_sz = src.size[0]*src.size[1]*sizeof(float);

    float B[] = { 1, 1, 1,
                  1, 1, 1,
                  1, 1, 1 };
    size_t bytesB = 9*sizeof(float);

    // Copy A and B^T into device memory
    //pb_SwitchToTimer( &timers, pb_TimerID_COMPUTE );

    //showInOut(src, E);
    Mat in, out;
    //resize(src, in, Size(HEIGHT, WIDTH));
    //resize(L, out, Size(HEIGHT, WIDTH));
    //imshow(input_window, in);
    //imshow(output_window, out);
    //waitKey(0);

    //NUM_FRAMES = 20;
    //pb_SwitchToTimer( &timers, visc_TimerID_COMPUTATION );
    struct InStruct* args = (struct InStruct*)malloc (sizeof(InStruct));
    // Check if the total elements is a multiple of block size
    int block_x = 256;
    assert(src.size[0]*src.size[1] % block_x == 0);

    //imshow(input_window, src);
    //imshow(output_window, E);
    //waitKey(0);
    NUM_FRAMES = 600;
    cap = VideoCapture(params->inpFiles[0]);
    getNextFrame(cap, src);
    llvm_visc_track_mem(src.data, I_sz);
    llvm_visc_track_mem(B, bytesB);
    llvm_visc_track_mem(D.data, I_sz);
    llvm_visc_track_mem(E.data, I_sz);
    llvm_visc_track_mem(L.data, I_sz);
    packData(args, (float*)src.data, I_sz,
                   B, bytesB,
                   (float*)D.data, I_sz,
                   (float*)E.data, I_sz,
                   (float*)L.data, I_sz,
                   src.size[0], src.size[1]
                   );


    pb_SwitchToTimer( &timers, visc_TimerID_COMPUTATION );
    for(int j=0; j<NUM_FRAMES; j++) {
      //std::cout << "Run: " << j << "\n";
      void* DFG = __visc__launch(0, LaplacianEstimate, (void*)args);
      __visc__wait(DFG);
    }

    pb_SwitchToTimer( &timers, pb_TimerID_NONE );

    llvm_visc_request_mem(L.data, I_sz);

    //resize(src, in, Size(HEIGHT, WIDTH));
    //resize(L, out, Size(HEIGHT, WIDTH));
    //imshow(output_window, out);
    //imshow(input_window, in);
    //waitKey(0);

    llvm_visc_untrack_mem(src.data);
    llvm_visc_untrack_mem(B);
    llvm_visc_untrack_mem(D.data);
    llvm_visc_untrack_mem(E.data);
    llvm_visc_untrack_mem(L.data);

    pb_PrintTimerSet(&timers);
    __visc__cleanup();

    pb_FreeParameters(params);

    return 0;

}
