/***************************************************************************
 *cr
 *cr            (C) Copyright 2010 The Board of Trustees of the
 *cr                        University of Illinois
 *cr                         All Rights Reserved
 *cr
 ***************************************************************************/

/* 
 * Main entry of dense matrix-matrix multiplication kernel
 */

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <malloc.h>
#include <vector>
#include <iostream>
#include <fstream>
#include <CL/cl.h>
#include <parboil.h>

// I/O routines
extern bool readColMajorMatrixFile(const char *fn, int &nr_row, int &nr_col, std::vector<float>&v);
extern bool writeColMajorMatrixFile(const char *fn, int, int, std::vector<float>&);
extern char* readFile(const char*);

// Parameters of tile sizes
//#define TILE_SZ 16
//#define VEC_SZ 8

#define TILE_N 8
#define TILE_TB_HEIGHT 8
#define TILE_M (TILE_N*TILE_TB_HEIGHT)

#define CHECK_ERROR(errorMessage)           \
  if(clStatus != CL_SUCCESS)                \
  {                                         \
     std::cout<< errorMessage <<": "<< clStatus <<" Error!\n";  \
     std::cout<<"Line: "<<__LINE__<<"\n";   \
     exit(1);                               \
  }

void basicSgemm( char transa, char transb, int m, int n, int k, float alpha, cl_mem A, int lda, cl_mem B, int ldb, float beta, cl_mem C, int ldc, cl_kernel clKernel, cl_command_queue clCommandQueue, pb_TimerSet& timers )
{
  if ((transa != 'N') && (transa != 'n')) {
    std::cerr << "unsupported value of 'transa' in regtileSgemm()" << std::endl;
    return;
  }
  
  if ((transb != 'T') && (transb != 't')) {
    std::cerr << "unsupported value of 'transb' in regtileSgemm()" << std::endl;
    return;
  }
  
  // In this code we assume the matrix sizes are multiple of tile size
  //if ((m%TILE_SZ) || (n%TILE_SZ)) {
    //std::cerr << "unsupported size of matrix. m should be multiple of " << TILE_SZ
      //<< "; n should be multiple of " << TILE_SZ << std::endl;
  //}

  //size_t db[2] = {TILE_SZ/VEC_SZ,TILE_SZ};
  //size_t dg[2] = {m/TILE_SZ*db[0],n/TILE_SZ*db[1]};

  size_t dg[2] = {m*TILE_N/TILE_M,n*TILE_TB_HEIGHT/TILE_N};
  size_t db[2] = {TILE_N,TILE_TB_HEIGHT};
  cl_int clStatus;
 
  clStatus = clSetKernelArg(clKernel,0,sizeof(cl_mem),(void*)&A);
  clStatus = clSetKernelArg(clKernel,1,sizeof(int),(void*)&lda);
  clStatus = clSetKernelArg(clKernel,2,sizeof(cl_mem),(void*)&B);
  clStatus = clSetKernelArg(clKernel,3,sizeof(int),(void*)&ldb);
  clStatus = clSetKernelArg(clKernel,4,sizeof(cl_mem),(void*)&C);
  clStatus = clSetKernelArg(clKernel,5,sizeof(int),(void*)&ldc);
  clStatus = clSetKernelArg(clKernel,6,sizeof(int),(void*)&k);
  clStatus = clSetKernelArg(clKernel,7,sizeof(float),(void*)&alpha);
  clStatus = clSetKernelArg(clKernel,8,sizeof(float),(void*)&beta);
  CHECK_ERROR("clSetKernelArg")

  pb_SwitchToTimer(&timers, visc_TimerID_COMPUTATION);
  for(int i=0; i<1; i++) {
  
    clStatus = clEnqueueNDRangeKernel(clCommandQueue,clKernel,2,NULL,dg,db,0,NULL,NULL);
    CHECK_ERROR("clEnqueueNDRangeKernel")

    clStatus = clFinish(clCommandQueue); 
    CHECK_ERROR("clFinish")
  }
  pb_SwitchToTimer(&timers, pb_TimerID_NONE);
}

int main (int argc, char *argv[]) {

  struct pb_Parameters *params;
  struct pb_TimerSet timers;

  size_t A_sz, B_sz, C_sz;
  int matArow, matAcol;
  int matBrow, matBcol;
  std::vector<float> matA, matBT;


  /* Read command line. Expect 3 inputs: A, B and B^T 
     in column-major layout*/
  params = pb_ReadParameters(&argc, argv);
  if ((params->inpFiles[0] == NULL) 
      || (params->inpFiles[1] == NULL)
      || (params->inpFiles[2] == NULL)
      || (params->inpFiles[3] != NULL))
    {
      fprintf(stderr, "Expecting three input filenames\n");
      exit(-1);
    }

  /* Read in data */
  // load A
  readColMajorMatrixFile(params->inpFiles[0],
      matArow, matAcol, matA);
  // load B^T
  readColMajorMatrixFile(params->inpFiles[2],
      matBcol, matBrow, matBT);

  pb_InitializeTimerSet(&timers);

  pb_SwitchToTimer(&timers, visc_TimerID_SETUP);
  cl_int clStatus;

  cl_uint numPlatforms; 
  clStatus = clGetPlatformIDs(0, NULL, &numPlatforms);

  cl_platform_id clPlatform[numPlatforms];
  clStatus = clGetPlatformIDs(numPlatforms, clPlatform, NULL);
  CHECK_ERROR("clGetPlatformIDs")

  //char buffer[1000];
  //size_t bytes;
  //clStatus = clGetPlatformInfo(clPlatform[1], CL_PLATFORM_EXTENSIONS, 1000, buffer, &bytes);
  //CHECK_ERROR("clGetPlatformInfo")

  //printf("\nExtensions: %s\n", buffer);

  cl_device_id clDevice;
  clStatus = clGetDeviceIDs(clPlatform[1],CL_DEVICE_TYPE_CPU,1,&clDevice,NULL);
  CHECK_ERROR("clGetDeviceIDs")

  //printf("Device id = %p\n", clDevice);
  //cl_device_partition_property props[3];
  //props[0] = CL_DEVICE_PARTITION_EQUALLY;
  //props[1] = 1;
  //props[2] = 0;
  //cl_device_id subdevice_id[8];
  //cl_uint num_entries = 8;

  //cl_uint numDevices;
  //clCreateSubDevices(clDevice, props, num_entries, subdevice_id, &numDevices);
  //printf("Num of devices = %d\n", numDevices);
  //for(unsigned i =0 ; i< numDevices; i++) {
    //printf("Subdevice id %d = %p\n", i, subdevice_id[i]);
  //}
  //clDevice = subdevice_id[0];
 
  cl_context_properties clCps[3] = {CL_CONTEXT_PLATFORM,(cl_context_properties)clPlatform[1],0};
  //cl_context clContext = clCreateContextFromType(clCps,CL_DEVICE_TYPE_CPU,NULL,NULL,&clStatus);
  cl_context clContext = clCreateContext(clCps, 1, &clDevice, NULL, NULL, &clStatus);
  CHECK_ERROR("clCreateContextFromType")
   

  cl_command_queue clCommandQueue = clCreateCommandQueue(clContext,clDevice,CL_QUEUE_PROFILING_ENABLE,&clStatus);
  CHECK_ERROR("clCreateCommandQueue")

  pb_SetOpenCL(&clContext, &clCommandQueue);

  //size_t binarySize = 1112;
  //std::cout << "Size of binary = " << binarySize << "\n";
  //cl_program clProgram = clCreateProgramWithBinary(clContext, 1, &clDevice, &binarySize, (const unsigned char**)clSource, NULL, &clStatus);
  //cl_program clProgram = clCreateProgramWithSource(clContext,1,clSource,NULL,&clStatus);
  cl_kernel clKernel;
  cl_program clProgram;
  pb_CreateAndBuildKernelFromBinary("kernel.ir", "mysgemmNT", &clContext, &clDevice, &clProgram, &clKernel);
  CHECK_ERROR("Binary")
  //cl_program clProgram = clCreateProgramWithSource(clContext,1,clSource,NULL,&clStatus);
  //CHECK_ERROR("clCreateProgramWithSource")

  //char clOptions[50];
  //sprintf(clOptions,"");

  //clStatus = clBuildProgram(clProgram,1,&clDevice,clOptions,NULL,NULL);
  //CHECK_ERROR("clBuildProgram")



  //cl_kernel clKernel = clCreateKernel(clProgram,"mysgemmNT",&clStatus);
  //CHECK_ERROR("clCreateKernel")

  pb_SwitchToTimer( &timers, pb_TimerID_COMPUTE );
  // copy A to device memory
  A_sz = matArow*matAcol*sizeof(float);
  B_sz = matBrow*matBcol*sizeof(float);

  // allocate space for C
  C_sz = matArow*matBcol*sizeof(float);

  // OpenCL memory allocation
  std::vector<float> matC(matArow*matBcol);
  
  pb_SwitchToTimer( &timers, pb_TimerID_COPY );
  cl_mem dA = clCreateBuffer(clContext,CL_MEM_READ_ONLY,A_sz,NULL,&clStatus);
  CHECK_ERROR("clCreateBuffer")
  cl_mem dB = clCreateBuffer(clContext,CL_MEM_READ_ONLY,B_sz,NULL,&clStatus);
  CHECK_ERROR("clCreateBuffer")
  cl_mem dC = clCreateBuffer(clContext,CL_MEM_WRITE_ONLY,C_sz,NULL,&clStatus);
  CHECK_ERROR("clCreateBuffer")

  // Copy A and B^T into device memory
  pb_SwitchToTimer( &timers, pb_TimerID_COPY );
  clStatus = clEnqueueWriteBuffer(clCommandQueue,dA,CL_FALSE,0,A_sz,&matA.front(),0,NULL,NULL);
  CHECK_ERROR("clEnqueueWriteBuffer")
  clStatus = clEnqueueWriteBuffer(clCommandQueue,dB,CL_FALSE,0,B_sz,&matBT.front(),0,NULL,NULL);
  CHECK_ERROR("clEnqueueWriteBuffer")

  for(int i=0;i<matC.size();i++)
	matC[i] = 0.0f;

  clStatus = clEnqueueWriteBuffer(clCommandQueue,dC,CL_TRUE,0,C_sz,&matC.front(),0,NULL,NULL);
  CHECK_ERROR("clEnqueueWriteBuffer")

  pb_SwitchToTimer( &timers, pb_TimerID_COMPUTE );
  pb_SwitchToTimer( &timers, pb_TimerID_KERNEL );

  // Use standard sgemm interface
  basicSgemm('N', 'T', matArow, matBcol, matAcol, 1.0f, \
      dA, matArow, dB, matBcol, 0.0f, dC, matArow, clKernel, clCommandQueue, timers);

  pb_SwitchToTimer( &timers, pb_TimerID_COPY );
  clEnqueueReadBuffer(clCommandQueue,dC,CL_TRUE,0,C_sz,&matC.front(),0,NULL,NULL);

  pb_SwitchToTimer( &timers, visc_TimerID_SETUP);
  clStatus = clReleaseKernel(clKernel);
  clStatus = clReleaseProgram(clProgram);
  clStatus = clReleaseMemObject(dA);
  clStatus = clReleaseMemObject(dB);
  clStatus = clReleaseMemObject(dC);
  clStatus = clReleaseCommandQueue(clCommandQueue);
  clStatus = clReleaseContext(clContext); 
 
  pb_SwitchToTimer(&timers, pb_TimerID_NONE);
  pb_PrintTimerSet(&timers);
  
  if (params->outFile) {
   
    /* Write C to file */
    //pb_SwitchToTimer(&timers, pb_TimerID_IO);
    writeColMajorMatrixFile(params->outFile,
	matArow, matBcol, matC); 
  }


  double GPUtime = pb_GetElapsedTime(&(timers.timers[pb_TimerID_KERNEL]));
  std::cout<< "GFLOPs = " << 2.* matArow * matBcol * matAcol/GPUtime/1e9 << std::endl;
  pb_FreeParameters(params);

  //free((void*)clSource[0]);

 
  return 0;
}
