#!/usr/bin/python

import csv
import sys
import os
import collections
from optparse import OptionParser

import numpy as np
import matplotlib.pyplot as plt

default_value = '0.0'
num_figs = 0

# parse a csv file and create a dictionary holding all the data
def parseCSVFile(filename):
  # open file
  file = open(filename, "rb")

  # create the dictionary
  n = lambda k: None
  csvDict = collections.defaultdict(
              lambda: collections.defaultdict(
                lambda: collections.defaultdict(
                  lambda: collections.defaultdict(
                    lambda: collections.defaultdict(lambda: default_value)))))

  # parse file and create dict
  reader = csv.DictReader(file)
  try:
    applicationKey = ''
    versionKey = ''
    testKey = ''
    timerKey = ''
    categoryKey = ''
    for row in reader:
      if row['Application']:
        applicationKey = row['Application']
      if row['Version']:
        versionKey = row['Version']
      if row['Test']:
        testKey = row['Test']
      if row['Timer']:
        timerKey = row['Timer']
      if row['Category']:
        categoryKey = row['Category']
      csvDict[applicationKey][versionKey][testKey][timerKey][categoryKey] = row['Time in seconds']
  except csv.Error, e:
    sys.exit('file %s, line %d: %s' % (filename, reader.line_num, e))

  # close file
  file.close()

  # check correctness
  file = open(filename, "rb")
  reader = csv.DictReader(file)
  try:
    applicationKey = ''
    versionKey = ''
    testKey = ''
    timerKey = ''
    categoryKey = ''
    for row in reader:
      if row['Application']:
        applicationKey = row['Application']
      if row['Version']:
        versionKey = row['Version']
      if row['Test']:
        testKey = row['Test']
      if row['Timer']:
        timerKey = row['Timer']
      if row['Category']:
        categoryKey = row['Category']
      assert(csvDict[applicationKey][versionKey][testKey][timerKey][categoryKey] == row['Time in seconds'])
  except csv.Error, e:
    sys.exit('file %s, line %d: %s' % (filename, reader.line_num, e))
  file.close()

  #print csvDict['a']['b']['c']['d']['e']
  #print csvDict['sgemm']['visc']['c']['d']['e']
  #print csvDict['sgemm']['opencl_base']['c']['d']['e']
  #print csvDict['sgemm']['opencl_base']['small']['d']['e']
  #print csvDict['sgemm']['opencl_base']['small']['Final']['e']
  #print csvDict['sgemm']['opencl_base']['small']['Final']['IO']

  #for app in csvDict.iterkeys():
  #  for v in csvDict[app].iterkeys():
  #    for test in csvDict[app][v].iterkeys():
  #      for t in csvDict[app][v][test].iterkeys():
  #        for cat in csvDict[app][v][test][t].iterkeys():
  #          time = csvDict[app][v][test][t][cat]
  #          print (app, v, test, t, cat, time)

  # return dict
  return csvDict


# returns a list of available tests for the given application
# the tests are found based on the visc version, because it exists
# for all apps in the dict
def getTests(app, csvDict):
  return csvDict[app]["visc"].keys()


def isViscVersion(version):
  return version.startswith("visc")


def getAllVersions(csvDict):
  apps = csvDict.keys()

  versions = set()
  for app in apps:
    versions = versions | set(csvDict[app].keys())

  versionsList = sorted(list(versions))
  return versionsList


# print total execution time for all applications and corresponding
# test sizes
def printTotalExecutionTimeTable(csvDict):
  versions = getAllVersions(csvDict)

  line = "application"
  for v in versions:
    line = line + "," + v
  print line

  t = 'Final'
  cat = 'Timer Wall - IO'
  for app in csvDict.iterkeys():
    tests = getTests(app, csvDict)
    for test in tests:
      line = "{0:s}-{1:s}".format(app, test)
      for v in versions:
        line = line + "," + csvDict[app][v][test][t][cat]
      print line


def printTimerDecomposition(csvDict, version):
  # get apps
  apps = csvDict.keys()

  isVisc = isViscVersion(version)

  # get tests for each app
  tests = dict()
  for app in apps:
    tests[app] = csvDict[app][version].keys()

  # list of timer-category pairs
  if isVisc:
    timers =[('Final', 'Kernel'), 
             ('Final', 'Load Program Binary'), 
             ('Final', 'Argument Unpack'), 
             ('Final', 'Marshal Arguments'), 
             ('Final', 'Free Memory'), 
             ('Final', 'Memory Track'), 
             ('Final', 'Clear Context'), 
             ('Final', 'Total GPU Computation'), 
             ('Final', 'Copy Pointer Arguments'), 
             ('Final', 'Initialize Context'), 
             ('Final', 'Read Output'), 
             ('Final', 'Pthread Create'), 
             ('Final', 'Copy Scalar Arguments'), 
             ('Final', 'WorkGroup Size Calculation'), 
             ('Final', 'Output Pack'), 
             ('Parboil', 'Mem_Untrack'), 
             ('Parboil', 'Clear_Ctx'),
             ('Final', 'Timer Wall - IO'),
             ('Final', 'IO'), 
             ('GenVISC_Timer', 'Timer Wall Time')]
  else: 
    timers =[('Final', 'Init_Ctx'),
             ('Final', 'Arg_Unpack'), 
             ('Final', 'Copy_Scalar'), 
             ('Final', 'Mem_Track'), 
             ('Final', 'Driver'), 
             ('Final', 'Output_Unpack'), 
             ('Final', 'Arg_Pack'), 
             ('Final', 'Copy'), 
             ('Final', 'Compute'), 
             ('Final', 'Setup'), 
             ('Final', 'Read_Output'), 
             ('Final', 'Pthread_Create'), 
             ('Final', 'Kernel'), 
             ('Final', 'Mem_Free'), 
             ('Final', 'Copy Async'), 
             ('Final', 'Copy_Ptr'),
             ('Final', 'Output_Pack'),
             ('Final', 'Mem_Untrack'),
             ('Final', 'Clear_Ctx'),
             ('Final', 'Timer Wall - IO'),
             ('Final', 'IO'), 
             ('Final', 'Timer Wall Time')]

  print version
  line = "Category"
  for app in apps:
    for test in tests[app]:
      line = line + "," + app + "-" + test
  print line
 
  for (t, cat) in timers:
    line = cat
    for app in apps:
      for test in tests[app]:
        line = line + "," + csvDict[app][version][test][t][cat]
    print line

# command line options parser
parser = OptionParser()
parser.add_option("-f","--file",action="store",type="string",dest="filename")
parser.add_option("--print-totals",action="store_true",default=False,dest="printTotals")
parser.add_option("--print-timers",action="store",type="string",dest="printTimers")
parser.add_option("--print-all",action="store_true",default=False,dest="printAll")

# main
def main():
  global num_figs
  # parse command line arguments
  (options, args) = parser.parse_args()

  # find out input
  if not options.filename and not args:
    print "parboilReader.py: Error: No input file was given!"
    raise OSError
  elif not options.filename:
    filename = args[0]
  else:
    filename = options.filename

  # create a dict for the csv file
  csvDict = parseCSVFile(filename)

  # print tables
  if options.printTotals or options.printAll:
    printTotalExecutionTimeTable(csvDict)
    print ''

  if options.printTimers == "all" or options.printAll:
    versions = getAllVersions(csvDict)
    for v in versions:
      printTimerDecomposition(csvDict, v)
    print ''
  elif options.printTimers:
    printTimerDecomposition(csvDict, options.printTimers)
    print ''


if __name__ == '__main__':
  main()
